import abc
from email.message import EmailMessage


class Signer(abc.ABC):  # pragma: no cover
    @abc.abstractmethod
    def sign(self, message: EmailMessage) -> EmailMessage:
        raise NotImplementedError()



from .base import Transport
from .console import ConsoleTransport
from .file import FileTransport
from .memory import InMemoryTransport
from .multi import MultiTransport
from .null import NullTransport
from .stream import StreamTransport

__all__ = [
    "Transport",
    "ConsoleTransport",
    "FileTransport",
    "InMemoryTransport",
    "StreamTransport",
    "NullTransport",
    "MultiTransport",
]



import typing
from email.message import EmailMessage

from mailers import create_transport_from_url
from mailers.encrypters import Encrypter
from mailers.exceptions import DeliveryError, InvalidSenderError
from mailers.message import Email, Recipients
from mailers.preprocessors import Preprocessor
from mailers.signers import Signer
from mailers.transports import Transport

if typing.TYPE_CHECKING:  # pragma: nocover
    import jinja2
else:  # pragma: nocover
    try:
        import jinja2
    except ImportError:
        jinja2 = None


class Mailer:
    """A facade for sending mails."""

    def __init__(
        self,
        transport: typing.Union[Transport, str],
        from_address: typing.Optional[str] = None,
        signer: typing.Optional[Signer] = None,
        encrypter: typing.Optional[Encrypter] = None,
        preprocessors: typing.Optional[typing.List[Preprocessor]] = None,
    ) -> None:
        if isinstance(transport, str):
            transport = create_transport_from_url(transport)
        self.transport = transport
        self.from_address = from_address
        self.signer = signer
        self.encrypter = encrypter
        self.preprocessors = preprocessors or []

    async def send(self, message: typing.Union[Email, EmailMessage]) -> None:
        from_ = message.from_address if isinstance(message, Email) else message.get("From")
        sender_ = message.sender if isinstance(message, Email) else message.get("Sender")

        if not from_ and not self.from_address and not sender_:
            raise InvalidSenderError('Message must have "From" or "Sender" header.')

        if not from_ and self.from_address:
            if isinstance(message, Email):
                message.from_address = self.from_address
            else:
                message["From"] = self.from_address

        mime_message = message.build() if isinstance(message, Email) else message

        for preprocessor in self.preprocessors:
            mime_message = preprocessor(mime_message)

        if self.encrypter:
            mime_message = self.encrypter.encrypt(mime_message)

        if self.signer:
            mime_message = self.signer.sign(mime_message)

        try:
            await self.transport.send(mime_message)
        except Exception as ex:
            raise DeliveryError("Failed to deliver email message.") from ex

    async def send_message(
        self,
        to: Recipients,
        subject: str,
        text: typing.Any = None,
        html: typing.Any = None,
        from_address: typing.Optional[Recipients] = None,
        cc: typing.Optional[Recipients] = None,
        bcc: typing.Optional[Recipients] = None,
        reply_to: typing.Optional[Recipients] = None,
        headers: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        sender: typing.Optional[str] = None,
        return_path: typing.Optional[str] = None,
        message_id: typing.Optional[str] = None,
    ) -> None:
        message = Email(
            to=to,
            subject=str(subject or ""),
            from_address=from_address,
            cc=cc,
            bcc=bcc,
            reply_to=reply_to,
            headers=dict(headers) if headers else {},
            sender=sender,
            text=str(text) if text is not None else None,
            html=str(html) if html is not None else None,
            return_path=return_path,
            text_charset="utf-8",
            html_charset="utf-8",
            message_id=message_id,
        )
        await self.send(message)


class TemplatedMailer(Mailer):
    def __init__(
        self,
        transport: typing.Union[Transport, str],
        jinja_env: "jinja2.Environment",
        **kwargs: typing.Any,
    ) -> None:
        super().__init__(transport, **kwargs)
        self.jinja_env = jinja_env

    async def send_templated_message(
        self,
        to: Recipients,
        subject: str,
        text_template: typing.Optional[str] = None,
        html_template: typing.Optional[str] = None,
        template_context: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        from_address: typing.Optional[Recipients] = None,
        cc: typing.Optional[Recipients] = None,
        bcc: typing.Optional[Recipients] = None,
        reply_to: typing.Optional[Recipients] = None,
        headers: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        sender: typing.Optional[str] = None,
        return_path: typing.Optional[str] = None,
        message_id: typing.Optional[str] = None,
    ) -> None:
        assert text_template or html_template, "Either text_template or html_template must be set."
        text: typing.Optional[str] = None
        html: typing.Optional[str] = None
        if text_template:
            text = self.jinja_env.get_template(text_template).render(template_context or {})
        if html_template:
            html = self.jinja_env.get_template(html_template).render(template_context or {})

        message = Email(
            to=to,
            subject=str(subject or ""),
            from_address=from_address,
            cc=cc,
            bcc=bcc,
            reply_to=reply_to,
            headers=dict(headers) if headers else {},
            sender=sender,
            text=text,
            html=html,
            return_path=return_path,
            text_charset="utf-8",
            html_charset="utf-8",
            message_id=message_id,
        )
        await self.send(message)



from email.message import EmailMessage, MIMEPart

import css_inline


def _process(part: MIMEPart) -> MIMEPart:
    # https://github.com/Stranger6667/css-inline/tree/master/bindings/python
    inliner = css_inline.CSSInliner()

    content = part.get_content()
    part.set_content(
        # set_content requires a byte literal, not a string
        inliner.inline(content).encode("utf-8"),
        maintype="text",
        subtype="html",
    )
    return part


def css_inliner(message: EmailMessage) -> EmailMessage:
    """
    Search for any HTML part that is not attachment and inline CSS.

    Modified MIME part in-place.
    """
    if message.get_content_type() == "text/html":
        if not message["content-disposition"]:
            _process(message)

    if message.get_content_type() in [
        "multipart/alternative",
        "multipart/mixed",
        "multipart/related",
    ]:
        for part in message.get_payload():
            css_inliner(part)

    return message



from __future__ import annotations

from email.message import Message

from mailers.transports.base import Transport


class NullTransport(Transport):
    async def send(self, message: Message) -> None:
        pass



import sys
import typing

from mailers.transports.stream import StreamTransport


class ConsoleTransport(StreamTransport):
    def __init__(self, stream: typing.Literal["stdout", "stderr"] = "stderr") -> None:
        assert stream in ["stdout", "stderr"], "Unsupported console stream type: %s." % stream
        output = None
        if stream == "stderr":
            output = sys.stderr
        elif stream == "stdout":
            output = sys.stdout
        super().__init__(output)



import abc
from email.message import EmailMessage


class Encrypter(abc.ABC):  # pragma: no cover
    @abc.abstractmethod
    def encrypt(self, message: EmailMessage) -> EmailMessage:
        raise NotImplementedError()



import typing
from email.message import Message

from mailers.transports.base import Transport


class StreamTransport(Transport):
    def __init__(self, output: typing.IO) -> None:
        self.stream = output

    async def send(self, message: Message) -> None:
        self.stream.write(str(message))



from __future__ import annotations

import typing
from email.message import EmailMessage

from mailers.transports.base import Transport


class InMemoryTransport(Transport):
    def __init__(self, storage: typing.Optional[typing.List[EmailMessage]] = None) -> None:
        self.storage = [] if storage is None else storage

    @property
    def mailbox(self) -> typing.List[EmailMessage]:
        return self.storage

    async def send(self, message: EmailMessage) -> None:
        self.storage.append(message)



from __future__ import annotations

import typing
from email.message import Message

from mailers.transports.base import Transport


class SMTPTransport(Transport):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 25,
        user: typing.Optional[str] = None,
        password: typing.Optional[str] = None,
        use_tls: typing.Optional[bool] = None,
        timeout: int = 10,
        key_file: typing.Optional[str] = None,
        cert_file: typing.Optional[str] = None,
        cert_bundle: typing.Optional[str] = None,
        validate_certs: typing.Optional[bool] = None,
    ):
        self._host = host
        self._user = user
        self._port = port
        self._password = password
        self._use_tls = use_tls or False
        self._timeout = timeout
        self._key_file = key_file
        self._cert_file = cert_file
        self._cert_bundle = cert_bundle
        self._validate_certs = validate_certs if validate_certs is not None else True

    async def send(self, message: Message) -> None:
        import aiosmtplib

        sender = message.get("Sender")
        if sender:
            del message["Sender"]
        return_path = message.get("Return-Path")
        if return_path:
            del message["Return-Path"]
        if sender is None and return_path:
            sender = return_path

        await aiosmtplib.send(
            message,
            sender=sender,
            hostname=self._host,
            port=self._port,
            use_tls=self._use_tls,
            username=self._user,
            password=self._password,
            timeout=self._timeout,
            client_key=self._key_file,
            client_cert=self._cert_file,
            cert_bundle=self._cert_bundle,
            validate_certs=self._validate_certs,
        )



from __future__ import annotations

import abc
from email.message import EmailMessage


class Transport(abc.ABC):  # pragma: nocover
    @abc.abstractmethod
    async def send(self, message: EmailMessage) -> None:
        raise NotImplementedError()



import typing
from email.message import EmailMessage

from mailers.exceptions import MailersError
from mailers.transports.base import Transport


class MultiDeliveryError(MailersError):
    def __init__(self, message: str, exceptions: typing.List[Exception]) -> None:
        super().__init__(message)
        self.exceptions = exceptions


class MultiTransport(Transport):
    def __init__(self, transports: typing.Iterable[Transport]) -> None:
        self.transports = list(transports)

    async def send(self, message: EmailMessage) -> None:
        exceptions: typing.List[Exception] = []
        for transport in self.transports:
            try:
                await transport.send(message)
            except Exception as ex:
                exceptions.append(ex)
            else:
                return

        raise MultiDeliveryError("Failed to deliver message via configured mailers.", exceptions)



from __future__ import annotations

from dataclasses import dataclass

import anyio as anyio
import email
import email.encoders
import email.utils
import mimetypes
import os
import random
import string
import typing
from email.headerregistry import Address
from email.message import EmailMessage, Message
from email.mime.base import MIMEBase
from email.policy import SMTP

from mailers.exceptions import InvalidBodyError

if typing.TYPE_CHECKING:  # pragma: no cover
    from _typeshed import OpenBinaryMode, OpenTextMode

Recipients = typing.Union[str, Address, typing.Iterable[typing.Union[str, Address]]]


def _randon_string(length: int) -> str:
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=length))


def _string_to_address(value: typing.Union[str, Address]) -> Address:
    if isinstance(value, Address):
        return value

    if "<" in value:
        start_pos = value.index("<")
        end_pos = value.index(">")
        display_name = value[0 : start_pos - 1]
        email = value[start_pos + 1 : end_pos]
        return Address(display_name=display_name, addr_spec=email)
    return Address(addr_spec=value)


def _to_addresses(recipients: typing.Optional[Recipients] = None) -> typing.List[Address]:
    if recipients is None:
        return []
    if isinstance(recipients, str):
        return [_string_to_address(recipients)]
    if isinstance(recipients, Address):
        return [recipients]
    return list(map(_string_to_address, recipients))


AddressType = typing.Union[typing.Union[str, Address], typing.Iterable[typing.Union[str, Address]]]


class AddressList:
    def __init__(self, addresses: typing.Optional[AddressType] = None) -> None:
        self._addresses: typing.List[Address] = []
        if addresses is not None:
            if isinstance(addresses, (str, Address)):
                addresses = [addresses]
            self.set(*addresses)

    @property
    def empty(self) -> bool:
        return len(self) == 0

    @property
    def first(self) -> typing.Optional[Address]:
        try:
            return self._addresses[0]
        except IndexError:
            return None

    def set(self, *address: typing.Union[str, Address]) -> None:
        self._addresses = _to_addresses(address)

    def add(self, *address: typing.Union[str, Address]) -> None:
        self._addresses.extend(_to_addresses(address))

    def clear(self) -> None:
        self._addresses = []

    def __str__(self) -> str:
        return ", ".join(map(str, self._addresses))

    def __repr__(self) -> str:  # pragma: no cover
        return f'<AddressList: addresses="{self}">'

    def __iter__(self) -> typing.Iterator[Address]:
        return iter(self._addresses)

    def __len__(self) -> int:
        return len(self._addresses)

    def __bool__(self) -> bool:
        return not self.empty

    def __eq__(self, other: object) -> bool:
        return str(self) == str(other)

    def __get__(self, obj: object, type: typing.Optional[type] = None) -> AddressList:
        return self

    def __set__(self, obj: object, value: typing.Optional[Recipients]) -> None:
        self._addresses = _to_addresses(value)


def _sanitize_input(
    path: typing.Union[str, os.PathLike],
    name: typing.Optional[str] = None,
    content_type: typing.Optional[str] = None,
) -> typing.Tuple[str, str]:
    name = name or os.path.basename(path)
    content_type = content_type or mimetypes.guess_type(path)[0] or "application/octet-stream"
    return name, content_type


@dataclass
class Attachment:
    name: typing.Optional[str] = None
    content_type: typing.Optional[str] = None
    body: typing.Optional[typing.Union[str, bytes]] = None
    path: typing.Optional[str] = None
    inline: typing.Optional[bool] = False
    part: typing.Optional[Message] = None

    @property
    def mime_type_parts(self) -> typing.Tuple[str, str]:
        if self.content_type:
            main_type, sub_type = self.content_type.split("/")
            return main_type, sub_type
        return "application", "octet-stream"


class Email:
    to = AddressList()
    cc = AddressList()
    bcc = AddressList()
    reply_to = AddressList()
    from_address = AddressList()

    def __init__(
        self,
        to: typing.Optional[Recipients] = None,
        subject: typing.Optional[str] = None,
        text: typing.Optional[str] = None,
        html: typing.Optional[str] = None,
        from_address: typing.Optional[Recipients] = None,
        cc: typing.Optional[Recipients] = None,
        bcc: typing.Optional[Recipients] = None,
        reply_to: typing.Optional[Recipients] = None,
        headers: typing.Optional[dict] = None,
        sender: typing.Optional[str] = None,
        return_path: typing.Optional[str] = None,
        text_charset: str = "utf-8",
        html_charset: str = "utf-8",
        boundary: typing.Optional[str] = None,
        message_id: typing.Optional[str] = None,
    ) -> None:
        self._sender: typing.Optional[Address] = None
        self._attachments: typing.List[Attachment] = []

        self.to = to
        self.cc = cc
        self.bcc = bcc
        self.reply_to = reply_to
        self.from_address = from_address

        if sender:
            self.sender = _string_to_address(sender)
        self.return_path = return_path
        self.subject = subject
        self.html = html
        self.html_charset = html_charset
        self.text = text
        self.text_charset = text_charset
        self.boundary = boundary
        self.headers = headers or {}
        self.id = message_id

        self.date = email.utils.localtime()

    @property
    def sender(self) -> typing.Optional[Address]:
        return self._sender

    @sender.setter
    def sender(self, value: typing.Optional[typing.Union[str, Address]]) -> None:
        self._sender = _string_to_address(value) if value else None

    def attach(
        self,
        body: typing.Union[str, bytes],
        name: typing.Optional[str] = None,
        content_type: typing.Optional[str] = None,
    ) -> None:
        if isinstance(body, str):
            body = body.encode()
        self._attachments.append(Attachment(body=body, name=name, content_type=content_type))

    async def attach_from_path(
        self,
        path: typing.Union[str, os.PathLike],
        name: typing.Optional[str] = None,
        content_type: typing.Optional[str] = None,
        mode: typing.Union["OpenBinaryMode", "OpenTextMode"] = "r",
    ) -> None:
        name, content_type = _sanitize_input(path, name, content_type)
        async with await anyio.open_file(path, mode=mode) as f:
            self.attach(await f.read(), name, content_type)

    def attach_from_path_sync(
        self,
        path: typing.Union[str, os.PathLike],
        name: typing.Optional[str] = None,
        content_type: typing.Optional[str] = None,
        mode: str = "r",
    ) -> None:
        name, content_type = _sanitize_input(path, name, content_type)
        with open(path, mode) as f:
            self.attach(f.read(), name, content_type)

    def embed(
        self,
        body: typing.Union[str, bytes],
        name: typing.Optional[str] = None,
        content_type: typing.Optional[str] = None,
    ) -> None:
        self._attachments.append(Attachment(body=body, name=name, content_type=content_type, inline=True))

    async def embed_from_path(
        self,
        path: typing.Union[str, os.PathLike],
        name: typing.Optional[str] = None,
        content_type: typing.Optional[str] = None,
        mode: typing.Union["OpenTextMode", "OpenBinaryMode"] = "r",
    ) -> None:
        name, content_type = _sanitize_input(path, name, content_type)
        async with await anyio.open_file(path, mode) as f:
            self.embed(await f.read(), name, content_type)

    def embed_from_path_sync(
        self,
        path: typing.Union[str, os.PathLike],
        name: typing.Optional[str] = None,
        content_type: typing.Optional[str] = None,
        mode: str = "r",
    ) -> None:
        name, content_type = _sanitize_input(path, name, content_type)
        with open(path, mode) as f:
            self.embed(f.read(), name, content_type)

    def attach_part(self, part: MIMEBase) -> None:
        self._attachments.append(Attachment(part=part))

    def validate(self) -> None:
        if all([self.text is None, self.html is None, not self._attachments]):
            raise InvalidBodyError("Email message must have a text, or HTML part or attachments.")

    def build(self) -> EmailMessage:  # noqa: C901
        self.validate()

        if not self.id:
            domain = ""
            if self.sender:
                domain = self.sender.domain
            if self.from_address:
                first_address = self.from_address.first
                if first_address:
                    domain = first_address.domain
                else:  # pragma: no cover
                    raise ValueError("Could not read sender domain from From header.")
            self.id = email.utils.make_msgid(domain=domain)

        headers = {
            "From": self.from_address,
            "To": self.to,
            "Cc": self.cc,
            "Bcc": self.bcc,
            "Reply-To": self.reply_to,
            "Return-Path": self.return_path,
            "Sender": self.sender,
            "Message-ID": self.id,
            "Date": self.date,
            "Subject": self.subject,
            **self.headers,
        }
        inline_attachments = [a for a in self._attachments if a.inline and not a.part]
        attachments = [a for a in self._attachments if not a.inline and not a.part]
        extra_parts = [a for a in self._attachments if a.part]

        mime_message = EmailMessage(policy=SMTP)

        for header_name, header_value in headers.items():
            if not header_value:  # ignore empty headers
                continue
            mime_message[header_name] = header_value

        # this is text only message
        if self.text and not any([self.html, inline_attachments, attachments]):
            mime_message.set_content(self.text, subtype="plain", charset=self.text_charset)
            return mime_message

        # this is HTML only message
        if self.html and not any([self.text, inline_attachments, attachments]):
            mime_message.set_content(self.html, subtype="html", charset=self.html_charset)
            return mime_message

        if self.text:
            mime_message.set_content(self.text, subtype="plain", charset=self.text_charset)

        if self.html:
            domain = str(self.id).split("@").pop()
            mime_message.add_alternative(self.html, subtype="html", charset=self.html_charset)
            html_part = mime_message.get_payload(1 if self.text else 0)
            for inline_attachment in inline_attachments:
                main_type, sub_type = inline_attachment.mime_type_parts
                cid = inline_attachment.name or _randon_string(16) + "@" + domain

                kwargs = {}
                if isinstance(inline_attachment.body, str):
                    kwargs["subtype"] = "plain" if sub_type not in ["html", "plain"] else sub_type
                else:
                    kwargs["maintype"] = main_type
                    kwargs["subtype"] = sub_type

                html_part.add_related(
                    inline_attachment.body,
                    disposition="inline",
                    filename=inline_attachment.name,
                    cid=cid,
                    headers=[
                        "Content-ID: <%s>" % cid,
                        "X-Attachment-ID: %s" % cid,
                    ],
                    **kwargs,
                )

        # this is attachments only message
        for attachment in attachments:
            main_type, sub_type = attachment.mime_type_parts
            mime_message.add_attachment(
                attachment.body,
                maintype=main_type,
                subtype=sub_type,
                disposition="attachment",
                filename=attachment.name,
            )

        for extra_part in extra_parts:
            if extra_part.part:
                mime_message.attach(extra_part.part)

        return mime_message

    def __str__(self) -> str:  # pragma: no cover
        return str(self.build())



import typing
from email.message import EmailMessage

try:  # pragma: no cover
    import dkim
except ImportError:  # pragma: no cover
    raise ImportError(
        "Please install dkimpy (https://pypi.org/project/dkimpy/) library " "to sign messages with DKIM method."
    )

from mailers.signers.base import Signer


class DKIMSigner(Signer):
    def __init__(
        self,
        selector: str,
        private_key: typing.Optional[str] = None,
        private_key_path: typing.Optional[str] = None,
        headers: typing.Optional[typing.Iterable[str]] = None,
    ) -> None:
        assert private_key or private_key_path, 'Either "dkim_key" or "dkim_key_path" must be passed.'
        self.dkim_selector = selector
        self.dkim_key = private_key
        self.dkim_key_path = private_key_path
        self.headers = headers or ["From", "To", "Subject"]

    def sign(self, message: EmailMessage) -> EmailMessage:
        key = self.dkim_key or ""
        if self.dkim_key_path:
            # we read file once and then cache in this instance
            # so it shouldn't bottleneck the flow.
            # this is intentional, to avoid marking this method async
            # if it doesn't work for you then pass key using "private_key" argument.
            with open(self.dkim_key_path, "r") as f:
                key = f.read()
                self.dkim_key = key  # cache

        from_address = message["From"]
        sender_domain = from_address.split("@")[-1]

        signature = dkim.sign(
            message=message.as_bytes(),
            selector=self.dkim_selector.encode(),
            domain=sender_domain.encode(),
            privkey=key.encode(),
            include_headers=self.headers,
        )
        message.add_header("DKIM-Signature", signature[len("DKIM-Signature: ") :].decode().replace("\r\n", " "))
        return message



import typing_extensions

import pytest
import typing
from email.message import EmailMessage

Mailbox: typing_extensions.TypeAlias = typing.List[EmailMessage]


@pytest.fixture
def mailbox() -> Mailbox:  # pragma: nocover
    return []



class MailersError(Exception):
    """Base error class."""


class NotRegisteredTransportError(MailersError):
    """Raised when transport is not registered but retrieved."""


class InvalidSenderError(MailersError):
    """Raised when From or Sender header is invalid."""


class InvalidBodyError(MailersError):
    """Raised when the message is empty."""


class DeliveryError(MailersError):
    """Raised if transport fails to deliver the message."""



import typing
from urllib.parse import parse_qsl, urlparse

from mailers.exceptions import NotRegisteredTransportError
from mailers.transports import ConsoleTransport, FileTransport, InMemoryTransport, NullTransport, Transport


def create_transport_from_url(url: str) -> Transport:
    """Create a transport instance from URL configuration."""
    components = urlparse(url)
    protocol = components.scheme
    options = dict(parse_qsl(components.query))

    if protocol == "console":
        assert "stream" in options

        stream = options.get("stream")
        assert stream in ["stdout", "stderr"], '"stream" must be one of "stdout", "stderr"'
        return ConsoleTransport(stream=typing.cast(typing.Literal["stdout", "stderr"], stream))

    if protocol == "file":
        return FileTransport(directory=components.path)

    if protocol == "memory":
        return InMemoryTransport()

    if protocol == "null":
        return NullTransport()

    if protocol == "smtp":

        def _cast_to_bool(value: str) -> bool:
            return value.lower() in ["yes", "1", "on", "true"]

        timeout: typing.Union[str, int, None] = options.get("timeout", None)
        if timeout is not None:
            timeout = int(timeout)

        use_tls = _cast_to_bool(options.get("use_tls", ""))
        key_file: typing.Optional[str] = options.get("key_file", None)
        cert_file: typing.Optional[str] = options.get("cert_file", None)
        cert_bundle: typing.Optional[str] = options.get("cert_bundle", None)
        validate_certs = _cast_to_bool(options.get("validate_certs", ""))

        from mailers.transports.smtp import SMTPTransport

        return SMTPTransport(
            components.hostname or "localhost",
            components.port or 25,
            components.username,
            components.password,
            use_tls=use_tls,
            timeout=timeout or 10,
            key_file=key_file,
            cert_file=cert_file,
            cert_bundle=cert_bundle,
            validate_certs=validate_certs,
        )

    raise NotRegisteredTransportError(f"Don't know how to create transport for protocol '{protocol}'.")



import re
from email.message import EmailMessage, MIMEPart
from typing import Callable


def remove_with_re(html: str):
    # https://stackoverflow.com/questions/28208186/how-to-remove-html-comments-using-regex-in-python
    return re.sub("(<!--.*?-->)", "", html, flags=re.DOTALL)


def remove_with_bs4(html: str):
    from bs4 import BeautifulSoup, Comment

    soup = BeautifulSoup(html, "html.parser")

    # Find all comments
    comments = soup.find_all(string=lambda text: isinstance(text, Comment))

    # Remove each comment
    for comment in comments:
        comment.extract()

    # Return the modified HTML as a string
    return str(soup)


def update_part_inline(part: MIMEPart, callback: Callable) -> MIMEPart:
    content = part.get_content()
    updated_content = callback(content)

    # set_content requires a byte literal, not a string
    if isinstance(updated_content, str):
        updated_content = updated_content.encode("utf-8")
    elif not isinstance(updated_content, bytes):
        raise TypeError(f"Expected str or bytes, got {type(updated_content)}")

    part.set_content(
        updated_content,
        maintype="text",
        subtype="html",
    )

    return part


try:
    import bs4  # noqa

    removal_strategy = remove_with_bs4
except ImportError:
    removal_strategy = remove_with_re


def remove_html_comments(message: EmailMessage) -> EmailMessage:
    if message.get_content_type() == "text/html":
        if not message["content-disposition"]:
            update_part_inline(message, removal_strategy)

    if message.get_content_type() in [
        "multipart/alternative",
        "multipart/mixed",
        "multipart/related",
    ]:
        for part in message.get_payload():
            # recurse!
            remove_html_comments(part)  # type: ignore

    return message



from mailers.encrypters import Encrypter
from mailers.exceptions import MailersError
from mailers.factories import create_transport_from_url
from mailers.mailer import Mailer, TemplatedMailer
from mailers.message import Email
from mailers.preprocessors import Preprocessor
from mailers.signers import Signer
from mailers.transports import (
    FileTransport,
    InMemoryTransport,
    MultiTransport,
    NullTransport,
    StreamTransport,
    Transport,
)
from mailers.transports.smtp import SMTPTransport

__all__ = [
    "Transport",
    "InMemoryTransport",
    "SMTPTransport",
    "NullTransport",
    "FileTransport",
    "StreamTransport",
    "MailersError",
    "Email",
    "Preprocessor",
    "Signer",
    "Encrypter",
    "Mailer",
    "TemplatedMailer",
    "create_transport_from_url",
    "MultiTransport",
]



from __future__ import annotations

import anyio as anyio
import datetime
import os
from email.message import Message

from mailers.transports.base import Transport


class FileTransport(Transport):
    def __init__(self, directory: str):
        if directory is None or directory == "":
            raise ValueError('Argument "path" of FileTransport cannot be None.')

        self.directory = directory

    async def send(self, message: Message) -> None:
        file_name = "message_%s.eml" % datetime.datetime.today().isoformat()
        output_file = os.path.join(self.directory, file_name)
        async with await anyio.open_file(output_file, "wb") as f:
            await f.write(message.as_bytes())
